package icu.wjz.service.impl;

import icu.wjz.entity.dto.user.UserDTO;
import icu.wjz.entity.po.user.UserPO;
import icu.wjz.mapper.UserMapper;
import icu.wjz.service.UserService;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.binding.BindingException;
import org.springframework.aop.framework.AopContext;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.TransactionTemplate;

import javax.annotation.Resource;
import java.util.Objects;
import java.util.Optional;

@Slf4j
@Service
@Transactional
public class UserServiceImpl implements UserService {

    @Resource
    private UserServiceImpl userService;
    @Resource
    private UserMapper userMapper;

    @Override
    @Transactional
    public ResponseEntity<Object> a1(UserDTO body) {
        b1(body);
        return null;
    }


    private void b1(UserDTO body){
        /**
         * 此时b1中SQL是会被a1开启的事物所管理
         * 如果b1()抛出异常，则a1中执行的变更也会被回退。
         * 但如果抛出的不是RuntimeException 则事务不会回退。
         * 下面会解释。
         */

        userMapper.insert(body);

        userMapper.errorInsert(body);
    }

    @Override
    public ResponseEntity<Object> a2(UserDTO body) {
        b2(body);
        return null;
    }

    @Transactional
    private void b2(UserDTO body){
        /**
         * 此时b2虽然打了@Transactional标签 但事务仍然不会启用
         * 原因：@Transactional 无法作用在私有方法上(private)
         */
        userMapper.insert(body);

        userMapper.errorInsert(body);

    }


    @Override
    public ResponseEntity<Object> a3(UserDTO body) {
        b3(body);
        return null;
    }

    @Override
    @Transactional
    public ResponseEntity<Object> b3(UserDTO body) {
        /**
         * 此时b2虽然打了@Transactional标签，并且为public 但事务仍然不会启用
         * 原因：默认情况下，只有来自外部的方法调用才会被aOP代理所捕获，
         * 也就是类内部调用本类方法，并不会引起事务行为，即使加了@Transactional
         *
         * 补充：如果想要事务生效需要使用上面注释调用方式
         */
        userMapper.insert(body);

        userMapper.errorInsert(body);
        return null;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public ResponseEntity<Object> a4(UserDTO body) {

        b4(body); // 事务只会处理未被处理的异常

        return null;
    }

    public void b4(UserDTO body) {


        /**
         * 此时a4正常开启事务， 但是如果b4方法抛出Exception异常 事务将不会回退
         * @Transactional 默认情况下只会回退继承自RuntimeException的异常
         * 如果想回退Exception， 需要设置
         * @Transactional(rollbackFor = Exception.class)
         * 此时抛出Exception以及子类的异常，事务都将会回退
         *
         * 补充：抛出Error错误时，事务都将回退。
         */

        userMapper.insert(body);
        userMapper.errorInsert(body);
    }

    @Override
    public ResponseEntity<Object> a5(UserDTO body) {
         userService.b3(body); //使用容器对象调用。
        return null;
    }

    @Override
    @Transactional
    public ResponseEntity<Object> syn1(String name) {

        UserPO user = userMapper.select(name);

        if(Objects.isNull(user)){
            log.info("这个用户不存在");
            synchronized (user.getName().intern()){
                UserDTO body = new UserDTO();
                body.setName("诸葛大力");
                body.setAge(22);
                userMapper.insert(body);
            }
        }

        return new ResponseEntity<Object>("执行完成", HttpStatus.OK);
    }
}
